"""

Usage:
   iguard_netcli --ping --hostname=<hostname> [--username=<username>]  
                 [--password=<password>] --dde=<dde_id> --dst=<ip> 
                 [--src=<dde_ip>] [--time=<time>] [--do_not_fragment]
                 [--payload_size=<payload>]
   iguard_netcli --iperf --hostname=<hostname> [--username=<username>]
                 [--password=<password>] --dde=<dde_id> --dst=<ip> 
                 [--src=<dde_ip>][--buffer_len=<len>][--parallel=<num>]
                 [--time=<time>][--iperf_version=<ver>][--reverse]
   iguard_netcli --traceroute --hostname=<hostname> [--username=<username>]
                 [--password=<password>] --dde=<dde_id> --dst=<ip> 
                 [--src=<dde_ip>]
   iguard_netcli  --dig --hostname=<hostname> [--username=<username>] 
                 [--password=<password>] --dde=<dde_id> --dst=<ip>
                 [--time=<time>]
   iguard_netcli  --ethtool --hostname=<hostname> [--username=<username>] 
                 [--password=<password>] --dde=<dde_id> [--interface=<dev>]
   iguard_netcli --ifconfig --hostname=<hostname> [--username=<username>] 
                 [--password=<password>] --dde=<dde_id> [--interface=<dev>]
   iguard_netcli --mtu --hostname=<hostname> [--username=<username>] 
                 [--password=<password>] --dde=<dde_id>  --dst=<ip>
                 [--src=<dde_ip>]
   iguard_netcli  --bonds --hostname=<hostname> [--username=<username>]
                 [--password=<password>] --dde=<dde_id> [--interface=<dev>]
   iguard_netcli --resolv --hostname=<hostname> [--username=<username>] 
                 [--password=<password>] --dde=<dde_id> 
   iguard_netcli --hostname=<hostname> --username=<username> --password=<password> --save
   iguard_netcli --delete
   iguard_netcli -h | --help
   iguard_netcli --version


Options:
   -H hostname --hostname=<hostname>              DNS name/IP  of InfiniGuard
   -u username --username=<username>              Administrator's username
   -p password --password=<password>              Administrator's password
   -d dde_id, --dde=<dde_id>                      DDE ID (1/2)
   -P, --ping                                     Ping from DDE
   -T, --traceroute                               Traceroute from DDE
   -N, --nslookup                                 Nslookup from DDE
   -E, --ethtool                                  Ethtool on interface(s)
   -i, --ifconfig                                 Ifconfig on interface(s)
   -M, --mtu                                      Calculate MTU between DDE and IP address
   -B, --bonds                                    Display bonds information
   -R, --resolv                                   Show contents of resolv.conf
   -I, --iperf                                    Run iperf client from DDE to Iperf server
   -D ip, --dst=<ip>                              Destination IP
   -S dde_ip, --src=<dde_ip>                      DDE IP to use as source
   -t sec, --time=<sec>                           Time to run the command (in seconds)
   --buffer_len=<len>                             Iperf buffer length (in KB)
   --parallel=<num>                               Iperf number of parallel threads
   --iperf_version=<ver>                          Iperf version to use (1 or 3) [default: 3]
   --reverse                                      Iperf3 only: reverse test direction
   --interface=<dev>                              Network interface to use
   --delete                                       Delete all stored credentials
   --save                                         Save credentials for hostname
   -h, --help                                     Show this screen
   -V, --version                                  Show iguard_netcli version
"""
from __future__ import print_function
from .logsetup import UserLogger, setup_logging, has_warnings
from docopt import docopt
from .__version__ import __version__ as infiniguard_netcli_version
import json
import os
from infinisdk import InfiniBox, Q
from infinisdk.core.exceptions import APICommandFailed
import requests
import ssl
import sys
import base64
import lzma
import arrow
from time import ctime, mktime, gmtime, asctime
import stat
from infi.credentials_store import CLICredentialsStore, Credentials
import pickle
from time import sleep
import traceback
from types import SimpleNamespace
import capacity
import sys
import ipaddress
from enum import Enum

requests.packages.urllib3.disable_warnings()

try:
    _create_unverified_https_context = ssl._create_unverified_context
except AttributeError:
    # Legacy Python that doesn't verify HTTPS certificates by default
    pass
else:
    # Handle target environment that doesn't support HTTPS verification
    ssl._create_default_https_context = _create_unverified_https_context


log = UserLogger(__name__)


class MyEnum(Enum):
    def __str__(self):
        return '%s' % " ".join(self.name.split("_")[1:])

    def __int__(self):
        return self.value

    def __eq__(self, other):
        return self.value == other

    def __rand__(self, other):
        return self.value & other

    @classmethod
    def find_one(cls, candidate):
        members = cls._member_names_
        for member in members:
            if candidate == cls[member]:
                return str(cls[member])

    @classmethod
    def find_all(cls, candidate):
        res = list()
        members = cls._member_names_
        for member in members:
            if candidate & cls[member]:
                res.append(str(cls[member]))
        return res


class IfconfigFlags(MyEnum):
    IFF_UP = 1 << 0
    IFF_BROADCAST = 1 << 1
    IFF_DEBUG = 1 << 2
    IFF_LOOPBACK = 1 << 3
    IFF_POINTOPOINT = 1 << 4
    IFF_NOTRAILERS = 1 << 5
    IFF_RUNNING = 1 << 6
    IFF_NOARP = 1 << 7
    IFF_PROMISC = 1 << 8
    IFF_ALLMULTI = 1 << 9
    IFF_MASTER = 1 << 10
    IFF_SLAVE = 1 << 11
    IFF_MULTICAST = 1 << 12
    IFF_PORTSEL = 1 << 13
    IFF_AUTOMEDIA = 1 << 14
    IFF_DYNAMIC = 1 << 15
    IFF_LOWER_UP = 1 << 16
    IFF_DORMANT = 1 << 17
    IFF_ECHO = 1 << 18


class XmitPolicy(MyEnum):
    POLICY_layer2 = 0
    POLICY_layer_3_4 = 1
    POLICY_layer_2_3 = 2
    POLICY_encapsulated_layer_2_3 = 3
    POLICY_encapsulated_layer_3_4 = 4


class AdSelect(MyEnum):
    AD_stable = 0
    AD_bandwidth = 1
    AD_count = 2


def get_function_name():
    return traceback.extract_stack(None, 2)[0][2]


def load_cookie(infinibox, username, hostname):
    try:
        with open('/tmp/iguard_netcli_cookie_{}_{}'.format(username, hostname), 'rb') as f:
            return pickle.load(f)
    except Exception:
        log.debug("Cookie not found.")
    return None


def save_cookie(infinibox, username, hostname):
    cookie = None
    try:
        with open('/tmp/iguard_netcli_cookie_{}_{}'.format(username, hostname), 'wb') as f:
            pickle.dump(infinibox.api._session.cookies.copy(), f)
    except Exception as e:
        log.debug("Failed to save cookie")
        log.debug(repr(e))


def my_login(infinibox, username, hostname):
    cookie = load_cookie(infinibox, username, hostname)
    if not cookie:
        log.debug("Cookie not found. Logging in.")
        infinibox.login()
        cookie = save_cookie(infinibox, username, hostname)
    else:
        log.debug("Loaded saved cookie")
        infinibox.api._session.cookies = cookie
        try:
            infinibox.get_version()
            # It didn't use the cookie, but looged in
            if infinibox.is_logged_in():
                log.debug("Logged in, cookie expired. Saving.")
                save_cookie(infinibox, username, hostname)
        except Exception:
            log.debug("Cookie not accepted")
            infinibox.api._session.cookies.clear()
            infinibox.login()
            save_cookie(infinibox, username, hostname)


class MyCLICrdentialsStore(CLICredentialsStore):
    def _get_file_folder(self):
        return ".infi.credentials_store"

    def authenticate(self, key, credentials):
        return True

    def ask_credentials_prompt(self, key):
        print('Connecting to system ' + str(key))


def delete_credentials(store):
    print("Deleting all credentials")
    store.reset_credentials()


def save_credentials(store, hostname, username, password):
    credentials = Credentials(username, password)
    store.set_credentials(hostname, credentials)
    print("Stored credentials for host {}".format(hostname))


def validate_config(arguments, store):
    config = {}
    if arguments['--help']:
        docopt(__doc__, '-h')

    for key, value in arguments.items():
        config[key[2:]] = value
    if not config['delete']:
        if not config['username'] or not config['password']:
            credentials = store.get_credentials(config['hostname'])
            config['username'] = credentials.get_username()
            config['password'] = credentials.get_password()
    return config


def is_task_finished(infinibox, dde_id, task_id):
    task_url = "/api/dde/{}/api/asynctasks/{}".format(dde_id, task_id)
    get = infinibox.api.get(task_url)
    status = get.get_result().get('status')
    return status != 'RUNNING'


def get_lines(infinibox, dde_id, task_id, last_id):
    result_url = "/api/dde/{}/api/asynctasks/{}/progress_results?sort=id&page_size=1000&page=1&id=gt:{}".format(
        dde_id, task_id, last_id)
    get = infinibox.api.get(result_url)
    res = get.get_result()
    return res


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_err(*args, **kwargs):
    print(f'{bcolors.FAIL}{" ".join([a for a in args])}{bcolors.ENDC}', file=sys.stderr, **kwargs)


def _print_lines(out):
    for line in out:
        if line.get('stdstream') == 2:
            print_err(line.get('line'))
        else:
            print(line.get('line'))


def get_command(config):
    return next((k for k, v in config.items() if v == True), None)


def resolv(config):
    dde_id = config['dde']
    infinibox = config['infinibox']
    get = infinibox.api.get(
        "/api/dde/{}/api/network/tools/resolv".format(dde_id))
    res = get.get_result()
    print('search', " ".join([a for a in res.get('search_domain')]))
    for nameserver in res.get('nameservers'):
        print('nameserver', nameserver)


def print_ifconfig(interface):
    i = SimpleNamespace(**interface)
    s = SimpleNamespace(**i.stats)
    flags = (i.flags if hasattr(i, 'flags') else 4099) % (1 << 16)
    flags_verbose = ",".join(IfconfigFlags.find_all(flags))
    hw_addr = i.phys_port_id if i.phys_port_id else i.hw_addr

    print(f'{i.ifname}: flags={flags}<{flags_verbose}>  mtu {i.mtu}')
    print(f'        ether: {hw_addr} txqueuelen {i.txqlen}  (Ethernet)')
    print(f'        RX packets {s.rx_packets}  bytes {s.rx_bytes} ({capacity.capacity.Capacity(s.rx_bytes * 8)})')
    print(f'        RX errors {s.rx_errors}  dropped {s.rx_dropped}  overruns {s.rx_over_errors}  crc {s.rx_crc_errors}')
    print(f'        TX packets {s.tx_packets}  bytes {s.tx_bytes} ({capacity.capacity.Capacity(s.tx_bytes * 8)})')
    print(f'        TX errors {s.tx_errors}  dropped {s.tx_dropped} carrier {s.tx_carrier_errors}  aborted {s.tx_aborted_errors}')
    print('')
    for alias in i.ip_aliases:
        a = SimpleNamespace(**alias)
        netmask = ipaddress.IPv4Interface(
            '{}/{}'.format(a.address, a.prefixlen)).netmask
        print(f'{a.intfname}: flags={flags}<{flags_verbose}>  mtu {i.mtu}')
        print(f'        inet: {a.address} netmask {netmask} broadcast {a.broadcast}')
        print(f'        ether: {hw_addr} txqueuelen {i.txqlen}  (Ethernet)')
        print('')


def print_ethtool(interface):
    i = SimpleNamespace(**interface)
    supported_modes = "\n                               ".join(
        [a for a in i.supported_modes]) if i.supported_modes else ''
    print(f'Settings for {i.ifname}:')
    print(f'        Supported ports: [ {i.supported_ports if i.supported_ports else ""} ]')
    print(f'        Supported link modes:   {supported_modes}')
    print(f'        Speed: {str(i.speed) + "Mb/s" if i.speed else "Unknown"}')
    print(f'        Duplex: {i.duplex if i.duplex else "Unknown"}')
    print(f'        Auto-negotiation: {"on" if i.autoneg else "off"}')
    print(f'        Link detected: {"yes" if i.operstate == "UP" else "no"}')
    print(f'        Link Failure Count: {i.carrier_changes}')
    print('')


def print_bond(bond, all_interfaces):
    bond_properties = bond.pop('bond_properties')
    ad_info = bond_properties.pop('bond_ad_info', None)
    i = SimpleNamespace(**bond)
    b = SimpleNamespace(**bond_properties)
    if b.bond_mode == 0:
        bond_mode = "load balancing (round-robin)"
    elif b.bond_mode == 1:
        bond_mode = "fault-tolerance (active-backup)"
    elif b.bond_mode == 4:
        bond_mode = "IEEE 802.3ad Dynamic link aggregation"
    # supported_modes
    print(f'Settings for {i.ifname}:')
    print(f'        Bonding Mode: {bond_mode}')
    if b.bond_mode == 4:
        print(f'        Transmit Hash Policy: {XmitPolicy.find_one(b.bond_xmit_hash_policy)} ({b.bond_xmit_hash_policy})')
    print(f'        MII Status: {"up" if i.carrier else "down"}')
    print(f'        MII Polling Interval (ms): {b.bond_miimon}')
    print(f'        Up Delay (ms): {b.bond_updelay}')
    print(f'        Down Delay (ms): {b.bond_downdelay}')
    if b.bond_mode == 4:
        a = SimpleNamespace(**ad_info)
        print('')
        print(f'        802.3ad info')
        print(f'        LACP rate: {"fast" if b.bond_ad_lacp_rate else "slow"}')
        print(f'        Min links: {b.bond_min_links}')
        print(f'        Aggregator selection policy (ad_select): {AdSelect.find_one(b.bond_ad_select)}')
        print(f'        System priority: {b.bond_ad_actor_sys_prio}')
        print(f'        System MAC address: {i.hw_addr}')
        print(f'        Active Aggregator Info:')
        print(f'                Aggregator ID: {a.bond_ad_info_aggregator}')
        print(f'                Number of ports: {a.bond_ad_info_num_ports}')
        print(f'                Actor Key: {a.bond_ad_info_actor_key}')
        print(f'                Partner Key: {a.bond_ad_info_partner_key}')
        print(f'                Partner Mac Address: {a.bond_ad_info_partner_mac}')

    for slave in b.bond_ports:
        interface = next(
            (a for a in all_interfaces if a.get('ifname') == slave), {})
        s = SimpleNamespace(**interface)
        print('')
        print(f'        Slave Interface: {s.ifname}')
        print(f'                MII Status: {"up" if s.carrier else "down"}')
        print(f'                Speed: {str(s.speed) + "Mb/s" if s.speed else "Unknown"}')
        print(f'                Duplex: {s.duplex if s.duplex else "Unknown"}')
        print(f'                Link detected: {"yes" if s.operstate == "UP" else "no"}')
        print(f'                Link Failure Count: {s.carrier_changes}')
        print(f'                Permanent HW addr: {s.phys_port_id if s.phys_port_id else s.hw_addr}')
    print('')


def get_interface_details(config):
    dde_id = config['dde']
    infinibox = config['infinibox']
    i = config.get('interface')
    if i:
        ifquery = '?name={}'.format(i)
    else:
        ifquery = ''
    get = infinibox.api.get(
        "/api/dde/{}/api/network/tools/interface_details{}".format(dde_id, ifquery))
    res = get.get_result()
    interfaces = [a for a in res if not a.get('backend_port')]
    return interfaces


def ifconfig(config):
    interfaces = get_interface_details(config)
    for interface in interfaces:
        print_ifconfig(interface)


def ethtool(config):
    interfaces = [a for a in get_interface_details(
        config) if a.get('kind') == 'port']
    interface_names = [a.get('ifname') for a in interfaces]
    interface_request = config.get('interface')
    if interface_request and interface_request not in interface_names:
        print_err(f'{interface_request} is not a physical interface.')
        sys.exit(1)
    for interface in interfaces:
        print_ethtool(interface)


def bonds(config):
    bonds = [a for a in get_interface_details(
        config) if a.get('kind') == 'bond']
    bond_names = [a.get('ifname') for a in bonds]
    bond_request = config.get('interface')
    if bond_request and bond_request not in bond_names:
        print_err(f'{bond_request} is not a bond interface.')
        sys.exit(1)
    config.pop('interface', None)
    all_interfaces = get_interface_details(config)
    for bond in bonds:
        print_bond(bond, all_interfaces)


def run_generic_command(config):
    log.info('Running {}'.format(get_function_name()))
    command = get_command(config)
    log.info("Command is {}".format(command))

    dde_id = config['dde']
    infinibox = config['infinibox']
    EXCLUDES = ['infinibox', 'version', 'username', 'password', 'dde']
    data = {k: v for k, v in config.items(
    ) if k not in EXCLUDES and v not in [None, False]}
    log.debug(data)
    post = infinibox.api.post(
        "/api/dde/{}/api/network/tools/{}".format(dde_id, command), data=data)

    task_id = post.get_result().get('task_id')
    task_url = "/api/dde/{}/api/asynctasks/{}".format(dde_id, task_id)
    last_id = 0
    try:
        while True:
            finished = is_task_finished(infinibox, dde_id, task_id)
            out = get_lines(infinibox, dde_id, task_id, last_id)
            if out:
                last_id = out[-1].get('id')
                _print_lines(out)

            if finished and not out:
                break
            else:
                sleep(0.1)
    finally:
        delete = infinibox.api.delete(task_url)
        log.debug('Delete status code: {}'.format(delete.status_code))


def infiniguard_netcli():
    try:
        log_handler = setup_logging()
        with log_handler.applicationbound():
            arguments = docopt(__doc__, version=infiniguard_netcli_version)
            store = MyCLICrdentialsStore("infiniguard")
            config = validate_config(arguments, store)
            hostname = config['hostname']
            username = config['username']
            password = config['password']
            dde = config['dde']
            command = get_command(config)
            if command not in ['save', 'delete']:
                infinibox = InfiniBox(hostname, use_ssl=True,
                                      auth=(username, password))
                my_login(infinibox, username, hostname)
                config['infinibox'] = infinibox
                get = infinibox.api.get('/api/dde/1/api/network/tools/resolv')
                if not get.response.content:
                    print_err('This system does not support iguard_netcli')
                    sys.exit(1)

            if command == 'delete':
                delete_credentials(store)
            elif command == 'save':
                save_credentials(store, hostname, username, password)
            elif command == 'ifconfig':
                ifconfig(config)
            elif command == 'ethtool':
                ethtool(config)
            elif command == 'bonds':
                bonds(config)
            elif command == 'resolv':
                resolv(config)
            else:
                run_generic_command(config)
    except APICommandFailed as e:
        if e.status_code == 404:
            print_err('This system does not support iguard_netcli')
        elif e.status_code == 503:
            print_err('System Unreachable')
        else:
            print_err(f'ERROR: {e.message}')
        sys.exit(1)

    except Exception as e:
        print(repr(e))
        sys.exit(1)
